/*
 * I am a driver for the DLI PC 8000 EPC. I control the device via its web
 * interface. I support querying the state (i.e. on or off) of the outlets and
 * turning each outlet on or off.
 *
 * Macro parameters:
 *   P     Prefix for PV names (e.g. "17bm:")
 *   D     Device name (e.g. "DliEpcr1" for DLI Ethernet Power Controller 1)
 *   M     Device model ID (PC8000 | EPCR2 | LPC2) (e.g. PC8000)
 *   AP    Asyn TCP/IP client port name for communicating with EPRC device
 *   USER  Username for authenticating with the device (e.g. "admin")
 *   PWD   Password for authenticating with the device (e.g. "open4me")
 *
 * Device model IDs:
 *   PC8000  PC-8000 Ethernet Power Controller
 *   EPCR2   Ethernet Power Controller II
 *   LPC2    Web Power Switch II
 */

program dliEpcr

option +r;

#include "dliEpcrLibSnlOrC.h"

%%#include <string.h>
%%#include <stdio.h>
%%#include <stdlib.h>

%%#include <epicsAssert.h>
%%#include <errlog.h>

%%#include <asynDriver.h>

%%#include <imca-asyn-utils.h>

%%#include "dliEpcrLib.h"

%{
  static short getOutletDesired(SS_ID ssId, struct UserVar *pVar, int outlet);
  static short basicGetOutletDesired(SS_ID ssId, struct UserVar *pVar,
      int outlet);
  static void logErrorMessage(SS_ID ssId, struct UserVar *pVar,
      const char *message);
}%

#define OUTLET_DESIRED_ON_VALUE_1 0
#define OUTLET_DESIRED_ON_VALUE_2 1
#define OUTLET_DESIRED_OFF_VALUE_1 2
#define OUTLET_DESIRED_OFF_VALUE_2 3

#define PV_PREFIX_KEY "P"
#define DEVICE_NAME_KEY "D"
#define DEVICE_MODEL_KEY "M"
#define DEVICE_ASYN_PORT_NAME_KEY "AP"
#define DEVICE_HTTP_USER_KEY "USER"
#define DEVICE_HTTP_PASSWORD_KEY "PWD"

char *DeviceModelAsString;
int DeviceModel;
char *DeviceAsynPortName;
char *DeviceAsynUser; /* Actual type: "asynUser *" */
char *DeviceHttpUser;
char *DeviceHttpPassword;

char ErrorMessage[1024];
int Index;
int OutletWithSwitchRequest;
int Status;
int snl_pvStatOK;

/* General control */

double PollInterval;
  assign PollInterval to "{P}{D}:PollIntrvl.RVAL";
  monitor PollInterval;
  evflag PollIntervalChanged;
  sync PollInterval to PollIntervalChanged;
double Refresh;
  assign Refresh to "{P}{D}:Refresh.PROC";
  monitor Refresh;
  evflag RefreshRequested;
  sync Refresh to RefreshRequested;

/* Outlet control */

short Outlet[NUM_OUTLETS];
  assign Outlet to {
    "{P}{D}:Out1Mbbi",
    "{P}{D}:Out2Mbbi",
    "{P}{D}:Out3Mbbi",
    "{P}{D}:Out4Mbbi",
    "{P}{D}:Out5Mbbi",
    "{P}{D}:Out6Mbbi",
    "{P}{D}:Out7Mbbi",
    "{P}{D}:Out8Mbbi"
  };

short Outlet1Desired;
  assign Outlet1Desired to "{P}{D}:Out1Mbbo";
  monitor Outlet1Desired;
  syncq Outlet1Desired 10;
short Outlet2Desired;
  assign Outlet2Desired to "{P}{D}:Out2Mbbo";
  monitor Outlet2Desired;
  syncq Outlet2Desired 10;
short Outlet3Desired;
  assign Outlet3Desired to "{P}{D}:Out3Mbbo";
  monitor Outlet3Desired;
  syncq Outlet3Desired 10;
short Outlet4Desired;
  assign Outlet4Desired to "{P}{D}:Out4Mbbo";
  monitor Outlet4Desired;
  syncq Outlet4Desired 10;
short Outlet5Desired;
  assign Outlet5Desired to "{P}{D}:Out5Mbbo";
  monitor Outlet5Desired;
  syncq Outlet5Desired 10;
short Outlet6Desired;
  assign Outlet6Desired to "{P}{D}:Out6Mbbo";
  monitor Outlet6Desired;
  syncq Outlet6Desired 10;
short Outlet7Desired;
  assign Outlet7Desired to "{P}{D}:Out7Mbbo";
  monitor Outlet7Desired;
  syncq Outlet7Desired 10;
short Outlet8Desired;
  assign Outlet8Desired to "{P}{D}:Out8Mbbo";
  monitor Outlet8Desired;
  syncq Outlet8Desired 10;

entry {
  %%pVar->snl_pvStatOK = pvStatOK;
  %%pVar->DeviceAsynUser = NULL;
}

ss Main {
  state initialize {
    when () {
      %%asynStatus status;

      DeviceModelAsString = macValueGet(DEVICE_MODEL_KEY);
      %%assert(pVar->DeviceModelAsString != NULL);
      if (strcmp("PC8000", DeviceModelAsString) == 0) {
        DeviceModel = DLI_MODEL_PC8000;
      } else if (strcmp("EPCR2", DeviceModelAsString) == 0) {
        DeviceModel = DLI_MODEL_EPCR2;
      } else if (strcmp("LPC2", DeviceModelAsString) == 0) {
        DeviceModel = DLI_MODEL_LPC2;
      } else {
        errlogPrintf("dliEpcr:error unknown model \"%s\"\n",
          DeviceModelAsString);
        assert(0);
      }
      DeviceAsynPortName = macValueGet(DEVICE_ASYN_PORT_NAME_KEY);
      %%assert(pVar->DeviceAsynPortName != NULL);
      DeviceHttpUser = macValueGet(DEVICE_HTTP_USER_KEY);
      %%assert(pVar->DeviceHttpUser != NULL);
      DeviceHttpPassword = macValueGet(DEVICE_HTTP_PASSWORD_KEY);
      %%assert(pVar->DeviceHttpPassword != NULL);

      %{
        status = imcaCreateIpClient((asynUser **)&pVar->DeviceAsynUser,
            pVar->DeviceAsynPortName, "", 0, "", 0);
      }%
      %%assert(status == asynSuccess);

      pvFreeQ(Outlet1Desired);
      pvFreeQ(Outlet2Desired);
      pvFreeQ(Outlet3Desired);
      pvFreeQ(Outlet4Desired);
      pvFreeQ(Outlet5Desired);
      pvFreeQ(Outlet6Desired);
      pvFreeQ(Outlet7Desired);
      pvFreeQ(Outlet8Desired);
      efClear(PollIntervalChanged);
      efClear(RefreshRequested);
    } state updateFromDevice
  }

  state watchForChanges {
    when (pvGetQ(Outlet1Desired)) {
      OutletWithSwitchRequest = 1;
    } state handleSwitchRequest
    when (pvGetQ(Outlet2Desired)) {
      OutletWithSwitchRequest = 2;
    } state handleSwitchRequest
    when (pvGetQ(Outlet3Desired)) {
      OutletWithSwitchRequest = 3;
    } state handleSwitchRequest
    when (pvGetQ(Outlet4Desired)) {
      OutletWithSwitchRequest = 4;
    } state handleSwitchRequest
    when (pvGetQ(Outlet5Desired)) {
      OutletWithSwitchRequest = 5;
    } state handleSwitchRequest
    when (pvGetQ(Outlet6Desired)) {
      OutletWithSwitchRequest = 6;
    } state handleSwitchRequest
    when (pvGetQ(Outlet7Desired)) {
      OutletWithSwitchRequest = 7;
    } state handleSwitchRequest
    when (pvGetQ(Outlet8Desired)) {
      OutletWithSwitchRequest = 8;
    } state handleSwitchRequest

    when (delay(PollInterval)) {
    } state handlePollIntervalDelayExpiration

    when (efTestAndClear(PollIntervalChanged)) {
    } state handlePollIntervalDelayExpiration

    when (efTestAndClear(RefreshRequested)) {
    } state handleRefreshRequest
  }

  state handleSwitchRequest {
    when () {
      %%int status;

      %{
        status = switchOutlet((asynUser *)pVar->DeviceAsynUser,
            pVar->DeviceModel, pVar->DeviceHttpUser, pVar->DeviceHttpPassword,
            pVar->OutletWithSwitchRequest,
            getOutletDesired(ssId, pVar, pVar->OutletWithSwitchRequest));
        if (status != 0) {
          sprintf(pVar->ErrorMessage, "failed to switch outlet %d",
            pVar->OutletWithSwitchRequest);
          logErrorMessage(ssId, pVar, pVar->ErrorMessage);
        }
      }%
    } state updateFromDevice
  }

  state handlePollIntervalDelayExpiration {
    when () {} state updateFromDevice
  }

  state handleRefreshRequest {
    when () {} state updateFromDevice
  }

  state updateFromDevice {
    when () {
      %{
        pVar->Status = getOutletStatesFromDevice(
            (asynUser *)pVar->DeviceAsynUser, pVar->DeviceModel,
            pVar->DeviceHttpUser, pVar->DeviceHttpPassword, pVar->Outlet);
      }%
      if (Status != 0) {
        for (Index = 0; Index < NUM_OUTLETS; Index++)
          Outlet[Index] = OUTLET_ERR;
        %{
          logErrorMessage(ssId, pVar,
              "failed to get outlet states from device");
        }%
      }
      for (Index = 0; Index < NUM_OUTLETS; Index++) {
        Status = pvPut(Outlet[Index]);
        if (Status != snl_pvStatOK) {
          sprintf(ErrorMessage, "failed to update Outlet%dBi from device",
            Index + 1);
          %%logErrorMessage(ssId, pVar, pVar->ErrorMessage);
        }
      }
    } state watchForChanges
  }
}

exit {
  %%imcaDestroyIpClient((asynUser *)pVar->DeviceAsynUser);
}

%{
static short getOutletDesired(SS_ID ssId, struct UserVar *pVar, int outlet) {
  short desiredState;

  desiredState = basicGetOutletDesired(ssId, pVar, outlet);
  return (desiredState == OUTLET_DESIRED_ON_VALUE_1 ||
    desiredState == OUTLET_DESIRED_ON_VALUE_2) ? OUTLET_ON : OUTLET_OFF;
}

static short basicGetOutletDesired(SS_ID ssId, struct UserVar *pVar,
    int outlet) {
  if (outlet == 1) return pVar->Outlet1Desired;
  if (outlet == 2) return pVar->Outlet2Desired;
  if (outlet == 3) return pVar->Outlet3Desired;
  if (outlet == 4) return pVar->Outlet4Desired;
  if (outlet == 5) return pVar->Outlet5Desired;
  if (outlet == 6) return pVar->Outlet6Desired;
  if (outlet == 7) return pVar->Outlet7Desired;
  if (outlet == 8) return pVar->Outlet8Desired;
  assert(0); return -1; /* Programmer error: outlet not between 1 and 8 */
}

static void logErrorMessage(SS_ID ssId, struct UserVar *pVar,
    const char *message) {
  asynPrint((asynUser *)pVar->DeviceAsynUser, ASYN_TRACE_ERROR,
    "%s:%s:error %s\n", pVar->DeviceAsynPortName, APPLICATION_NAME, message);
}
}%
