#!/usr/bin/python

#
# I simulate the DLI PC-8000 Ethernet Power Control device and its embedded
# HTTP server. I do not support all functions, just the ones that are useful
# for testing the EPICS "driver" for this device. From the status page (i.e.
# URI = "/" or "/index.shtml"), I support displaying the State (ON/OFF)
# column, Last switched column, and Action column (Switch ON/OFF). The Switch
# ON/OFF link works allowing an individual outlet to be switched on or off. I
# also support an extra function to reset the simulated device (i.e. switch
# all outlets on and make them all have the current time as their "last
# switched" date. This function is available by requesting the URI "/reset".
#

#
# Usage: sim-server.py [-d | --debug]
#
# Options:
#   -d/--debug   Print debug messages to stderr
#

import asynchat
import asyncore
import base64
import os
import re
import signal
import socket
import sys
import threading
import time

PORT = 50080

USERNAME = "admin"
PASSWORD = "admin"

IS_DEBUG_OUTPUT_ENABLED = False

def main(args):
  enableDebugOutputIfOptionPresent(args)
  ConnectionDispatcher(PORT)
  asyncore.loop()

def enableDebugOutputIfOptionPresent(args):
  if len(args) > 1 and args[1] in ["-d", "--debug"]:
    global IS_DEBUG_OUTPUT_ENABLED
    IS_DEBUG_OUTPUT_ENABLED = True

def initializeSignalHandlers():
  signalToNameMap = {
    signal.SIGHUP:"SIGHUP",
    signal.SIGINT:"SIGINT",
    signal.SIGPIPE:"SIGPIPE",
    signal.SIGALRM:"SIGALRM",
    signal.SIGTERM:"SIGTERM",
    signal.SIGUSR1:"SIGUSR1",
    signal.SIGUSR2:"SIGUSR2"
  }

  def signalHandler(signum, frame):
    if signum in [signal.SIGINT, signal.SIGTERM]:
      print
      print "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(0)
    else:
      print >> sys.stderr
      print >> sys.stderr, "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(1)

  for each in signalToNameMap.keys():
    signal.signal(each, signalHandler)

class ConnectionDispatcher(asyncore.dispatcher):
  def __init__(self, port):
    asyncore.dispatcher.__init__ (self)
    self.port = port
    self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
    self.set_reuse_addr()
    self.bind(("", port))
    self.listen(5)

  def handle_accept(self):
    ConnectionSession(self.accept())

class ConnectionSession(asynchat.async_chat):
  StatusPageGetRequestRe = re.compile("^GET /(?:index.shtml)? HTTP/[0-9.]+$")
  OnOrOffGetRequestRe = re.compile(
    r"^GET /outlet(on|off)\?([1-8]) HTTP/[0-9.]+$")
  ResetGetRequestRe = re.compile(r"^GET /reset HTTP/[0-9.]+$")

  def __init__(self, (conn, addr)):
    asynchat.async_chat.__init__(self, conn)
    self.buffer = ""
    self.set_terminator("\n")

    self.requestHeader = {}
    self.requestHeader["GET"] = ""
    self.requestHeader["Authorization"] = ""
    self.model = EthernetPowerControlDeviceModel.getInstance()

  def collect_incoming_data(self, data):
    self.buffer = self.buffer + data

  def found_terminator(self):
    data = self.buffer
    self.buffer = ""
    self.handleClientRequest(data)

  def push(self, data):
    asynchat.async_chat.push(self, data)
    if IS_DEBUG_OUTPUT_ENABLED: sys.stderr.write("> %s\n" % repr(data))

  def handleClientRequest(self, request):
    request = request.strip()
    if IS_DEBUG_OUTPUT_ENABLED: sys.stderr.write("< \"%s\"\n" % request)
    if request.startswith("GET "):
      self.requestHeader["GET"] = request
    elif request.startswith("Authorization: "):
      self.requestHeader["Authorization"] = request
    elif request == "":
      self.handleCompleteClientRequest()

  def handleCompleteClientRequest(self):
    if not self.isClientRequestAuthorized():
      self.sendUnauthorizedResponse()
    elif self.StatusPageGetRequestRe.match(self.requestHeader["GET"]):
      self.sendStatusPageResponse()
    elif self.OnOrOffGetRequestRe.match(self.requestHeader["GET"]):
      self.handleOnOrOffRequest()
    elif self.ResetGetRequestRe.match(self.requestHeader["GET"]):
      self.handleResetRequest()
    elif self.requestHeader["GET"].startswith("GET "):
      self.sendNotFoundResponse()
    self.close_when_done()

  def handleOnOrOffRequest(self):
    match = self.OnOrOffGetRequestRe.match(self.requestHeader["GET"])
    if match.group(1) == "off": self.model.turnOff(int(match.group(2)))
    else: self.model.turnOn(int(match.group(2)))
    self.sendSwitchedResponse()

  def handleResetRequest(self):
    self.model.reset()
    self.sendSwitchedResponse()

  def isClientRequestAuthorized(self):
    encodedCredentials = self.getEncodedAuthorizationCredentials()
    if encodedCredentials == None: return False
    try:
      credentials = base64.b64decode(encodedCredentials)
    except TypeError, e:
      return False
    usernameAndPassword = credentials.split(":", 1)
    if len(usernameAndPassword) != 2: return False
    return usernameAndPassword[0] == USERNAME and \
      usernameAndPassword[1] == PASSWORD

  def getEncodedAuthorizationCredentials(self):
    tokens = self.requestHeader["Authorization"].split(None, 2)
    if len(tokens) != 3: return None
    if tokens[1] != "Basic": return None
    return tokens[2]

  def sendUnauthorizedResponse(self):
    self.push("HTTP/1.0 401 Unauthorized\r\n")
    self.push("WWW-Authenticate: Basic realm=\"Password required\"\r\n")
    self.push("\r\n")

  def sendNotFoundResponse(self):
    self.push("HTTP/1.0 404 Not Found\r\n" + \
      "Server: EPC\r\n" + \
      "Connection: close\r\n" + \
      "Content-Type: text/html\r\n" + \
      "\r\n" + \
      "<HTML>\n" + \
      "<HEAD><TITLE>404 Not Found</TITLE></HEAD>\n" + \
      "<BODY><h1>Not Found</h1></BODY>\n" + \
      "</HTML>\n")

  def sendSwitchedResponse(self):
    self.push("HTTP/1.0 302 Found\r\n" + \
      "Location: /index.shtml\r\n" + \
      "Connection: close\r\n" + \
      "Content-Type: text/html\r\n" + \
      "\r\n" + \
      "<HTML>\n" + \
      "<HEAD><TITLE>document has moved</TITLE></HEAD>\n" + \
      "<BODY>\n" + \
      "document has moved to <A HREF=\"/index.shtml\">here</A>\n" + \
      "</BODY>\n" + \
      "</HTML>\n")

  def sendStatusPageResponse(self):
    self.push("HTTP/1.0 200 OK\r\n")
    self.push("Content-Type: text/html\r\n")
    self.push("Expires: Mon, 06 Jan 1990 00:00:01 GMT\r\n")
    self.push("Pragma: no-cache\r\n")
    self.push("Cache-Control: no-store, no-cache, must-revalidate, ")
    self.push("post-check=0, pre-check=0\r\n")
    self.push("\r\n")

    self.push(r"""<html>
<head>
<meta HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=us-ascii">
<META HTTP-EQUIV="Refresh" CONTENT="60">
<title>Outlet Control - Remote Power Controller</title></head>
<body alink="#0000DD" vlink="#0000DD">
<FONT FACE="Arial, Helvetica, Sans-Serif">
<table width="100%" cellspacing=0 cellpadding=0>
<tr>
<td valign=top width="17%" height="100%">

    <!-- menu -->
    <table width="100%" height="100%" align=center border=0 cellspacing=1
      cellpadding=0>
    <tr><td valign=top bgcolor="#F4F4F4">
    <table width="100%" cellpadding=1 cellspacing=5>

    <tr><td align=center>
    <table><tr><td>Digital Loggers, Inc.</td>
    <td><b><font size=-1>PC-8000 Ethernet Power Controller</font></b></td>
    </tr></table>
    <hr>
    </td></tr>

    <tr><td nowrap><b><a href="/index.shtml">Outlet Control</a></b></td></tr>
<tr><td nowrap><b><a href="/log.shtml">View log</a></b></td></tr>
<tr><td nowrap><b><a href="/admin.shtml">Settings</a></b></td></tr>
<tr><td nowrap><b><a href="/logout">Logout</a></b></td></tr>
<tr><td nowrap><b><a href="http://www.digital-loggers.com/pchelp.htm">Help</a>
  </b></td></tr>
<tr><td><hr></td></tr>
<tr><td><b><a href="www.google.com">GOOGLE</a></b></td></tr>
<tr><td><b><a href="http://www.digital-loggers.com">DLC</a></b></td></tr>
<tr><td><b><a href="www.yahoo.com">YAHOO</a></b></td></tr>
<tr><td><b><a href="www.CNN.com">CNN</a></b></td></tr>


    </table>
    </td></tr>

    <tr><td valign=bottom height="100%" bgcolor="#F4F4F4">
    <small> version 2.2.6 (Sep 23 2005 / 13:34:47)</small>
    </td></tr>

    </table>
    <!-- /menu -->

</td>
<td valign=top width="83%">

    <!-- heading table -->
    <table width="100%" align=center border=0 cellspacing=1 cellpadding=3>

        <tr>
        <th bgcolor="#DDDDFF" align=left>
        Controller: Remote Power Controller
        </th>
        </tr>

        <tr bgcolor="#FFFFFF" align=left>
        <td>
        System time: 09/18/2006 12:03:04
        </td>
        </tr>

    </table>
    <!-- /heading table -->

    <br>

    <!-- individual control table -->
    <table width="100%" align=center border=0 cellspacing=1 cellpadding=3>

        <tr>
        <th bgcolor="#DDDDFF" colspan=6 align=left>
        Individual Control
        </th>
        </tr>

        <!-- heading rows -->
        <tr bgcolor="#DDDDDD">
        <th>#</th>
        <th align=left>Name</th>
        <th align=left>State</th>
        <th align=left>Last switched</th>
        <th align=left colspan=2>Action</th>
        </tr>
        <!-- /heading rows -->

""")

    self.push("        %s\n" %  self.generateHtmlTableRowsForOutlets())

    self.push(r"""    </table>
    <!-- /individual control table -->

    <br>

        <table width="100%" align=center border=0 cellspacing=1 cellpadding=3>
  <tr><th bgcolor="#DDDDFF" align=left>Master Control</th></tr><tr>
  <td bgcolor="#F4F4F4" align=left><a href=outletoff?a>All outlets OFF</a>
  </td></tr><tr><td bgcolor="#F4F4F4" align=left>
  <a href=outleton?a>All outlets ON</a></td></tr><tr><td bgcolor="#F4F4F4"
  align=left><a href=outletgl?a>Cycle all outlets</a></td></tr><tr><td
  align=center>Sequence delay: 5 sec.</td></tr></table>

</td>
</tr>
</table>

</body>
</html>
""")

  def generateHtmlTableRowsForOutlets(self):
    buffer = ""
    for each in range(1,9):
      if each != 1: buffer = buffer + "\n        "
      buffer = buffer + self.generateHtmlTableRowForOutlet(each)
    return buffer

  def generateHtmlTableRowForOutlet(self, outlet):
    return ("<tr bgcolor=\"#F4F4F4\">" + \
      "<td align=center>%i</td>" + \
      "<td>Outlet Name %i</td>" + \
      "<td><b><font color=%s>%s</font></b></td>" + \
      "<td>%s&nbsp;</td>" + \
      "<td><a href=%s?%i>Switch %s</a></td>" + \
      "<td></td>" + \
      "</tr>") % \
        (outlet, outlet, self.getColorAttributeForOutletState(outlet),
         self.getOutletStateAsString(outlet),
         self.getLastSwitchedDateAsString(outlet),
         self.getSwitchUrlOptionName(outlet), outlet,
         self.getInvertedOutletStateAsString(outlet))

  def getColorAttributeForOutletState(self, outlet):
    if self.model.isOutletOn(outlet): return "green"
    else: return "red"

  def getInvertedOutletStateAsString(self, outlet):
    return self.basicGetOutletStateAsString(not self.model.isOutletOn(outlet))

  def getLastSwitchedDateAsString(self, outlet):
    return time.strftime("%m/%d/%Y %H:%M:%S",
      self.model.getOutletLastSwitchedDate(outlet))

  def getOutletStateAsString(self, outlet):
    return self.basicGetOutletStateAsString(self.model.isOutletOn(outlet))

  def basicGetOutletStateAsString(self, isOutletOn):
    if isOutletOn: return "ON"
    else: return "OFF"

  def getSwitchUrlOptionName(self, outlet):
    if self.model.isOutletOn(outlet): return "outletoff"
    else: return "outleton"

class EthernetPowerControlDeviceModel:
  SingletonLock = threading.RLock()
  Instance = None

  @classmethod
  def getInstance(cls):
    if cls.Instance != None: return cls.Instance
    cls.SingletonLock.acquire()
    if cls.Instance != None:
      cls.SingletonLock.release()
      return cls.Instance
    else:
      cls.Instance = EthernetPowerControlDeviceModel()
      cls.SingletonLock.release()
      return cls.Instance

  def __init__(self):
    self.outlets = {}
    self.outletLastSwitchedDates = {}
    self.reset()

  def getOutletLastSwitchedDate(self, outlet):
    return self.outletLastSwitchedDates[outlet]

  def isOutletOn(self, outlet):
    return self.outlets[outlet]

  def makeNowLastSwitchedDate(self, outlet):
    self.outletLastSwitchedDates[outlet] = time.localtime()

  def reset(self):
    for each in range(1,9): self.outlets[each] = True
    now = time.localtime()
    for each in range(1,9): self.outletLastSwitchedDates[each] = now

  def turnOff(self, outlet):
    if self.isOutletOn(outlet): self.makeNowLastSwitchedDate(outlet)
    self.outlets[outlet] = False

  def turnOn(self, outlet):
    if not self.isOutletOn(outlet): self.makeNowLastSwitchedDate(outlet)
    self.outlets[outlet] = True

# Bootstrap

if __name__ == '__main__':
  try:
    initializeSignalHandlers()
    main(sys.argv)
  except Exception, e:
    if isinstance(e, SystemExit):
      raise e
    else:
      print >> sys.stderr
      print >> sys.stderr, "Error: %s" % e
      sys.exit(1)
